# This file contains every functions needed to run the om_lhood_function
# Load every needed libraries

library(ape)
library(MASS)
# Data loading function
# param tree.nexus.path : the path of the nexus file on the disk
# param data.csv.path : the path of the csv file on the disk
# param
# return a tree, a regime and a edge matrix

load.OU.datas <- function(tree.nexus.path,regimes.csv.path,traits.csv.path){
	
	if(test.data.mode){
		tree <- read.nexus('test/tree.nex')
		tree$tip.label <- c(1,2,3,4)
		traits <- as.matrix(read.csv(traits.csv.path))[1:4,]	
		regimes_full_data <- as.matrix(read.csv('test/regimes.csv')) 
		edges <- regimes_full_data[,(length(regimes_full_data[1,])-2):(length(regimes_full_data[1,]))]
		regimes <- regimes_full_data[,1:(length(regimes_full_data[1,])-3)]
		plot(tree)
		nodelabels()
		return(list("tree"=tree,"regimes"=regimes,"edges"=edges,"traits"=traits) )
	}
	
	tree <- read.nexus(tree.nexus.path)
	traits <- as.matrix(read.csv(traits.csv.path))	
	regimes_full_data <- as.matrix(read.csv(regimes.csv.path)) 
	edges <- regimes_full_data[,(length(regimes_full_data[1,])-2):(length(regimes_full_data[1,]))]
	regimes <- regimes_full_data[,1:(length(regimes_full_data[1,])-3)]
	return(list("tree"=tree,"regimes"=regimes,"edges"=edges,"traits"=traits) )
	
	
}

# OU VCV transformation function
# params sigma.sq
# params alpha
# params t total length of the tree
# params s shared part of the tree (? not sure => check)
# return OU transformed vcv matrix

OU.vcv.transform <- function(sigma.sq,alpha,t,s){
	
	new.vcv <- sigma.sq * ( exp(-2*alpha*(t-s))*(1 - exp(-2*alpha*s))) / (2 * alpha) 
	
	return(new.vcv)
	
}

# NEW transform node age
OU.w.transform<-function(alpha,br.len){
	
	tr.br.len=exp(alpha*br.len)
	
	return(tr.br.len)
	
}

# NEW transform node age
OU.w.transform.total<-function(alpha,br.len){
	
	tr.br.len=exp(-alpha*br.len)
	
	return(tr.br.len)
	
}


# NEW function to get node age


# Creation of the VCV matrix
# params sigma.sq
# params alpha
# params phylo tree
# return OU vcv matrix

get.OU.vcv <- function(sigma.sq,alpha,tree){
	
	if(debug.mode){
		v <- vcv(tree)
		row.names(v) <- paste("", 1:nrow(v), sep="")
		colnames(v) <-  paste("" , 1:ncol(v), sep="")
		cat('\nStep 1 : vcv before its transformation (sp x sp) :\n','\n')
		print(v)
	}
	
	vcvmat <- OU.vcv.transform(sigma.sq,alpha,vcv(tree)[1],vcv(tree))
	#vcvmat <- vcv(tree)
	
	if(debug.mode){
		v <- vcvmat
		row.names(v) <- paste("", 1:nrow(v), sep="")
		colnames(v) <-  paste("" , 1:ncol(v), sep="")
		cat('\nStep 2 : vcv after the transformation (sp x sp)  :\n','\n')
		print(v)
	}
	
	return(vcvmat)
	
}

# Creation of the W matrix
# params regimes
# params edges
# params tree length
# params regimes number
# params species number
# return OU weigthed matrix

get.OU.w <- function(regimes,edges,tree.length,regimes.nb,species.nb,alpha){
	
	time <- as.matrix(branching.times(tree))
	age <- matrix(c(rep(0,species.nb)))	
	age <- rbind(age,time)
	
	transformed <- cbind(edges,vcv(tree)[1]-age[edges[,2]],vcv(tree)[1]-age[edges[,3]])
	
	transformed <- cbind(transformed,vcv(tree)[1])
	
	transformed <- cbind(transformed,OU.w.transform(alpha,transformed[,4]),OU.w.transform(alpha,transformed[,5]),OU.w.transform.total(alpha,transformed[,6]))
	
	transformed <- cbind(transformed,(transformed[,8]-transformed[,7])*transformed[,9])
	
	weighted.sum <- transformed[,10]
	
	if(debug.mode){
		cat('\nStep 3 : Transformation of column to obtain weighted sum ( = last column )\n','\n')
		print(transformed)
		cat('\nStep 4 : Regimes before transformation with weighted sum :\n','\n')
		print(regimes)		
	}
	
	# old ( false ) way to do
	#regimes <- regimes * (edges[,"branch_len"] / tree.length )
	
	# new way to do
	regimes <- regimes * weighted.sum
	
	if(debug.mode){
		cat('\nStep 5 : Regimes after transformation :\n','\n')
		print(regimes)
	}
	
	regimes.edges <- cbind(regimes,edges)
	
	w <- matrix(rep(0,regimes.nb*species.nb),ncol=species.nb)
	
	
	for(n in 1:species.nb){
		w<-get.regime.from.edge(n,n,w,regimes.edges,regimes.nb)
	}
	
	if(debug.mode){
		cat('\nStep 6 : W matrix (sp x rg) after the whole recursive process which retrive the edges values  :\n','\n')
		print(w)
	}
	
	regime0<-OU.w.transform.total(matrix(rep(tree.length,species.nb),ncol=species.nb),alpha)
	
	w <- rbind(regime0,w)
	
	if(debug.mode){
		cat('\nStep 7 : W with regime 0 (sp x rg)  :\n','\n')
		print(w)
	}
	
	
	return(w)
	
}

# Recursive function to retrive all regime informations for a specie by identifying the edges it belongs to.
# params node
# params specie
# params weighted matrix
# params regimes.edges
# params regime.nb
# return new weighted matrx

get.regime.from.edge <- function(node,specie,w,regimes.edges,regime.nb){
	
	edge <- regimes.edges[regimes.edges[,"to_node"]==node]
	
	if(length(edge) > 0 ){
		
		for(i in 1:regime.nb){
			w[i,specie] <- w[i,specie] + edge[i]				
		}
		
		w <- get.regime.from.edge(edge[regime.nb+2],specie,w,regimes.edges,regimes.nb)
		
	}	
	
	return(w)
	
}

OU.log.likelihood <- function(par,edges,regimes,tree,x,regimes.nb,species.nb,edge.nb,tree.length)		
{
	
	if(! debug.mode && ! verbose.mode){
		counter <<- counter + 1
		if(counter %% 10 == 0){
			cat('.')
		}
		if(counter %% 100 == 0){
			cat('.\n')
		}
	}
	
	alpha <- abs(par[1])
	sigma.sq <- abs(par[2])
	teta0 <- par[3]
	teta1 <- par[4]
	teta2 <- par[5]
	
	teta <- matrix(c(teta0,teta1,teta2),ncol=3)
	
	# Get the vcv
	vcv <- get.OU.vcv(alpha,sigma.sq,tree)
	
	# Get the W matrix
	w <- get.OU.w(regimes,edges,tree.length,regimes.nb,species.nb,alpha)  
		
	L <- -species.nb*log(2*pi)/2 - log(abs(det(vcv)))/2 + (-t(t(x)-t(w)%*%t(teta))%*%ginv(vcv)%*%(t(x)-t(w)%*%t(teta)))/2
	
	ret <- L[1,1]
	
	if(verbose.mode || debug.mode){
		cat('\nInfo : Params used : alpha:',alpha,', sigma.sq:',sigma.sq,', teta0:',teta0,', teta1:',teta1,', teta2:',teta2,'\n')
		
		cat('\nInfo : Log lik value : ',ret,'\n')
	}
	
	return(ret);
	
}